"""
:mod:`zsl.task.job_context`
---------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import annotations

from abc import abstractmethod
import contextlib
from threading import current_thread
from typing import Any, Callable, Generator, TypedDict

from flask.wrappers import Request, Response

from zsl.task.task_data import TaskData
from zsl.utils.reflection_helper import proxy_object_to_delegate


class JobData(TypedDict):
    """
    A dictionary that represents the data associated with a job.

    :ivar path: The path to the job data.
    :vartype path: str
    :ivar data: The data associated with the job.
    :vartype data: dict[str, Any]

    Example:
    --------
    >>> job_data = JobData(path='/path/to/job', data={'name': 'John', 'age': 30})
    """
    path: str
    data: dict[str, Any]


class Job:
    def __init__(self, data: JobData):
        self.data = data

    @property
    def path(self) -> str:
        """Job's path.

        :getter: Returns job's path
        :type: str
        """
        return self.data["path"]

    @property
    def payload(self) -> dict[str, Any]:
        """Data part of job.

        :getter: Returns job's payload
        :type: dict
        """
        return self.data["data"]

    @property
    def is_valid(self) -> bool:
        """Validity of job's data.

        :getter: Returns if job's data are valid
        :type: bool
        """
        return self.data and "path" in self.data and "data" in self.data


class JobContext:
    """Job Context"""

    def __init__(self, job: Job, task: object, task_callable: Callable):
        """
        Constructor
        """
        self._job = job
        self._task = task
        self._task_callable = task_callable
        self._task_data = TaskData(self.job.payload)

        JobContext._set_current_context(self)

    @property
    def job(self) -> Job:
        return self._job

    @property
    def task(self) -> object:
        return self._task

    @property
    def task_callable(self) -> Callable:
        return self._task_callable

    @property
    def task_data(self) -> TaskData:
        return self._task_data

    @classmethod
    def get_current_context(cls) -> JobContext:
        return current_thread()._current_job_context

    @classmethod
    def _set_current_context(cls, context: JobContext):
        current_thread()._current_job_context = context


class Responder:
    @abstractmethod
    def respond(self, r: Response) -> None:
        pass


class StatusCodeResponder(Responder):
    def __init__(self, status_code: int):
        self._status_code = status_code

    def respond(self, r: Response) -> None:
        r.status_code = self._status_code


def add_responder(responder: Responder) -> None:
    jc = JobContext.get_current_context()
    if isinstance(jc, WebJobContext):
        jc.add_responder(responder)


class HeadersDict(TypedDict):
    Location: str


class ResponseDict(TypedDict):
    headers: HeadersDict
    status_code: int


def web_task_redirect(location) -> HeadersDict:
    return {"headers": {"Location": location}, "status_code": 301}


def create_job(path: str, data: dict[str, Any]) -> Job:
    return Job({"data": data, "path": path})


class WebJobContext(JobContext):
    def __init__(
        self,
        path: str,
        data: dict,
        task: object,
        task_callable: Callable,
        request: Request,
    ):
        """Constructor"""
        super().__init__(create_job(path, data), task, task_callable)
        self._request = request
        self._responders: list[Responder] = []

    def get_web_request(self) -> Request:
        return self._request

    def add_responder(self, r: Responder):
        self._responders.append(r)

    def notify_responders(self, response: Response):
        try:
            for r in self._responders:
                r.respond(response)
        finally:
            self._responders = []


class DelegatingJobContext(JobContext):
    def __init__(self, job: Job, task: object, task_callable: Callable):
        wrapped_job_context = JobContext.get_current_context()
        super().__init__(job, task, task_callable)
        self._wrapped_job_context = wrapped_job_context
        proxy_object_to_delegate(self, wrapped_job_context)

    def stop_delegating(self):
        JobContext._set_current_context(self._wrapped_job_context)


@contextlib.contextmanager
def delegating_job_context(
    job: Job, task: object, task_callable: Callable
) -> Generator[DelegatingJobContext, None, None]:
    djc = DelegatingJobContext(job, task, task_callable)
    yield djc
    djc.stop_delegating()



"""
Test http responses of resource guard. It should return a model if policy is
met and an proper http code when policy is broken.
"""
import http.client
from unittest.case import TestCase

from mocks import mock

from zsl import Zsl
from zsl.application.containers.web_container import WebContainer
from zsl.interface.resource import ResourceResult
from zsl.resource.guard import Access, GuardedMixin, PolicyViolationError, ResourcePolicy, guard
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.http import HTTPTestCase, json_loads
from zsl.testing.test_utils import parent_module

TEST_VALUE_CREATED = 'created'
TEST_VALUE_READ = 'read'
TEST_VALUE_UPDATED = 'updated'
TEST_VALUE_DELETED = 'deleted'

SAMPLE_MODEL = {
    'test': 'value',
    'one': 1
}


class GuardedResourceTestResource:
    """This will be patched."""
    pass


class GuardedResourceTest(TestCase, HTTPTestCase):
    PATH = '/resource/guarded_resource_test'
    RESOURCE_CLASS = 'resource.guarded_resource_test.GuardedResourceTestResource'

    def setUp(self):
        config_object = IN_MEMORY_DB_SETTINGS.copy()
        # add this package as resource package for zsl to find the
        # `JsonServerModelResourceResource`
        config_object['RESOURCE_PACKAGES'] = ('resource',)
        zsl = Zsl(__name__, config_object=config_object,
                  modules=WebContainer.modules())
        zsl.testing = True

        # mock http requests
        self.app = zsl.test_client()

    def testAllowPolicy(self):
        class AllowPolicy(ResourcePolicy):
            default = Access.ALLOW

        @guard([AllowPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')

            self.assertHTTPStatus(
                http.client.OK,
                rv.status_code,
                "should return 200 status, returned data {0}".format(rv.data)
            )
            data = json_loads(rv.data)
            self.assertDictEqual(SAMPLE_MODEL, data,
                                 "should return sample model")

    def testDefaultDenyPolicy(self):
        class DenyPolicy(ResourcePolicy):
            default = Access.DENY

        @guard([DenyPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.FORBIDDEN, rv.status_code,
                                  "should return default 403 status")

    def testCustomDenyException(self):
        class PaymentError(PolicyViolationError):
            def __init__(self):
                super().__init__('Payment required', code=402)

        class PaymentPolicy(ResourcePolicy):
            @property
            def default(self):
                raise PaymentError()

        @guard([PaymentPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.PAYMENT_REQUIRED, rv.status_code,
                                  "should return custom status 402")

    def testCustomErrorHandling(self):
        def error_handler(*_):
            return ResourceResult(body={}, status=301)

        class DenyPolicy(ResourcePolicy):
            default = Access.DENY

        @guard(
            policies=[DenyPolicy()],
            exception_handlers=[error_handler]
        )
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.MOVED_PERMANENTLY, rv.status_code,
                                  "should return custom status 301")



"""
:mod:`zsl.tasks.asl.schedule_gearman_task`
------------------------------------------
"""
from zsl.task.task_data import TaskData
from zsl.utils.gearman_helper import schedule_gearman_task


class ScheduleGearmanTask:

    def perform(self, data):
        # type: (TaskData)->str
        data = data.payload

        # Create gearman.
        schedule_gearman_task(data['path'], data['data'])

        return 'job submitted'



"""
:mod:`zsl.application.service_application`
------------------------------------------

The module contains the main Zsl class - :class:`.ServiceApplication`. It is
responsible for gluing everything together.

There is a simple profile support. Profile is usually tied to a configuration
file for an environment. Usually environments slightly differ in settings -
especially connection strings. These twists can be managed via additional
configuration files with the `cfg` extension usually placed in the `settings`
package. The name of the environment configuration file is controlled via
`ZSL_SETTINGS` environment variable.


.. data:: SETTINGS_ENV_VAR_NAME

    Name of the environment variable to be read for the profile configuration
    file.

"""
import os
from typing import Any, Callable

from flask import Config, Flask
from injector import Binder, Injector, singleton

from zsl._state import set_current_app
from zsl.application.initialization_context import InitializationContext
from zsl.utils.warnings import deprecated
from zsl.version import version

# Name of the environment variable to be read for the profile configuration.
SETTINGS_ENV_VAR_NAME = 'ZSL_SETTINGS'


def get_settings_from_profile(profile, profile_dir=None):
    # type: (str, Any)->str
    """"Returns the  configuration file path for the given profile.

    :param profile: Profile name to be used.
    :param profile_dir: The directory where the profile configuration file should reside. It
                        may be also a module, and then the directory of the module is used.
    :return: Configuration file path.
    """
    if profile_dir is None:
        import settings
        profile_dir = settings

    if hasattr(profile_dir, '__file__'):
        profile_dir = os.path.dirname(profile_dir.__file__)

    return os.path.join(profile_dir, '{0}.cfg'.format(profile))


def set_profile(profile):
    # type (str)->None
    """Sets the current profile.

    :param profile: The profile to be used."""
    os.environ[SETTINGS_ENV_VAR_NAME] = get_settings_from_profile(profile)


class ServiceApplication(Flask):
    """Atteq Service Flask application."""

    VERSION = version

    def __init__(
        self,
        import_name,
        static_url_path=None,
        static_folder="static",
        static_host=None,
        host_matching=False,
        subdomain_matching=False,
        template_folder="templates",
        instance_path=None,
        instance_relative_config=False,
        root_path=None,
        modules=None,
        version=None,
        config_object=None,
        default_settings_module='settings.default_settings'
    ):
        super().__init__(
            import_name,
            static_url_path,
            static_folder,
            static_host,
            host_matching,
            subdomain_matching,
            template_folder,
            instance_path,
            instance_relative_config,
            root_path
        )
        self._dependencies_initialized = False
        self._default_settings_module = default_settings_module
        self._is_initialized = False
        self._injector = None
        self._worker = None
        self._configure(config_object)
        self._app_version = version
        if not modules:
            from zsl.application.containers.core_container import CoreContainer
            modules = CoreContainer.modules()
        self._configure_injector(modules)
        self._initialize()
        self._dependencies_initialized = True

    def __str__(self):
        return "ZSL(application={0}, zsl_version={1}, app_version={2})".format(
            self.name, self.VERSION,
            self._app_version)

    def _configure(self, config_object=None):
        # type: (Any) -> None
        """Read the configuration from config files.
        Loads the default settings and the profile settings if available.
        Check :func:`.set_profile`.

        :param config_object:
            This parameter is the configuration decscription may be a dict or
            string describing the module from which the configuration is used.
            Default is settings.default_settings.
        """

        if config_object:
            self.config.from_mapping(config_object)
        else:
            self.config.from_object(self._default_settings_module)

        zsl_settings = os.environ.get(SETTINGS_ENV_VAR_NAME)
        if zsl_settings is not None:
            self.config.from_envvar(SETTINGS_ENV_VAR_NAME)

    def _initialize(self):
        """Run the initializers."""
        ctx = self.injector.get(InitializationContext)
        ctx.initialize()

    def _register(self):
        """Register the current instance into application stack."""
        set_current_app(self)

    def _get_app_module(self):
        # type: () -> Callable
        """Returns a module which binds the current app and configuration.

        :return: configuration callback
        :rtype: Callable
        """

        def configure(binder):
            # type: (Binder) -> Callable
            binder.bind(ServiceApplication, to=self, scope=singleton)
            binder.bind(Config, to=self.config, scope=singleton)

        return configure

    def _configure_injector(self, modules):
        """Create the injector and install the modules.

        There is a necessary order of calls. First we have to bind `Config` and
        `Zsl`, then we need to register the app into the global stack and then
        we can install all other modules, which can use `Zsl` and `Config`
        injection.

        :param modules: list of injection modules
        :type modules: list
        """
        self._register()
        self._create_injector()
        self._bind_core()
        self._bind_modules(modules)
        self.logger.debug("Injector configuration with modules {0}.".format(modules))
        self._dependencies_initialized = True

    @deprecated
    def get_initialization_context(self):
        return self.injector.get(InitializationContext)

    def is_initialized(self):
        return self._dependencies_initialized

    @property
    def injector(self):
        # type: () -> Injector
        return self._injector

    @injector.setter
    def injector(self, value):
        self._injector = value

    def get_version(self):
        return self.version

    @property
    def version(self):
        app_version = self.app_version
        if app_version is None:
            return self.zsl_version
        else:
            return self.zsl_version + ":" + app_version

    @property
    def zsl_version(self):
        return ServiceApplication.VERSION

    @property
    def app_version(self):
        return self._app_version

    def _create_injector(self):
        self.logger.debug('Creating injector')
        self._injector = Injector()

    def _bind_core(self):
        self._injector.binder.bind(ServiceApplication, self, singleton)
        self._injector.binder.bind(Config, self.config, singleton)

    def _bind_modules(self, modules):
        for module in modules:
            self.injector.binder.install(module)



import unittest.mock as mock

from zsl.service.service import SessionFactory
from zsl.testing.db import TestSessionFactory as DbTestTestSessionFactory
from zsl.utils.injection_helper import bind


def mock_db_session():
    mock_sess = mock.MagicMock()

    def session_holder():
        return mock_sess

    class TestSessionFactory(DbTestTestSessionFactory):
        def __init__(self):
            super().__init__()
            self._session_holder = session_holder

    bind(SessionFactory, to=TestSessionFactory)
    bind(DbTestTestSessionFactory, to=TestSessionFactory)
    return mock_sess



"""
:mod:`zsl.interface.celery.worker`
----------------------------------

Implementation of celery workers.

.. moduleauthor:: Peter Morihladko
"""
from celery import Celery, shared_task

from zsl import Config, Injected, inject
from zsl.interface.task_queue import JobResult, TaskQueueWorker
from zsl.task.job_context import Job, JobData


class CeleryTaskQueueWorkerBase(TaskQueueWorker):
    """Base class for celery task queue worker.

    It contains only the task execution part of worker.
    """

    def execute_celery_task(self, job_data: dict) -> JobResult:
        """Creates job from task and executes the job.

        :param job_data: job's data
        :return: job results
        :rtype: dict
        """
        job = Job(job_data)
        return self.execute_job(job)


class CeleryTaskQueueOutsideWorker(CeleryTaskQueueWorkerBase):
    """Celery worker used only for task execution.

    This can be used when the worker is managed with some other application,
    like `celery worker` or `celery multi`.
    """

    def stop_worker(self):
        self._app.logger.error("Running from celery worker, kill from shell!")

    def run(self):
        self._app.logger.error("Running from celery worker, start from shell!")


@inject(config=Config)
def create_celery_app(config: Config = Injected) -> Celery:
    celery_app = Celery()
    celery_app.config_from_object(config["CELERY"])
    return celery_app


class CeleryTaskQueueMainWorker(CeleryTaskQueueWorkerBase):
    """Worker implementation for Celery task queue."""

    def __init__(
        self,
    ):
        super().__init__()
        self.celery_app = create_celery_app()

    def stop_worker(self):
        self._app.logger.error(
            "This is a celery app worker, kill the instance to stop it."
        )

    def run(self, argv: list[str]):
        """
        Run the celery worker cmd with given arguments from the list.

        Note: the first argument should be "worker".
        """
        self._app.logger.info("Running the worker.")
        self.celery_app.worker_main(argv)


@shared_task
@inject(worker=CeleryTaskQueueWorkerBase)
def zsl_task(job_data: JobData, worker: CeleryTaskQueueWorkerBase = Injected) -> JobResult:
    """
    Executes a task registered with Celery using the provided job data.

    `job_data` is a dictionary that describes the path to the desired task along with the payload. Specifically,
    it should contain a 'path' key pointing to the task and a 'data' key with the payload as a dictionary.

    :param job_data: A dictionary containing the path to the task and its payload.
    :type job_data: dict
    :param worker: The Celery worker responsible for executing the task. *Injected.*
    :type worker: CeleryTaskQueueWorkerBase
    :return: The result of the executed task.
    :rtype: JobResult

    :Example:
    >>> job_data = {
            "path": "some_module/some_task_function",
            "data": {
                "param1": "value1",
                "param2": "value2"
            }
        }
    """
    return worker.execute_celery_task(job_data)



"""
:mod:`zsl.application.error_handler`
---------------------------------------------------

This module does the error handling. It allows users to register
an error handler for a given exception type. It also provides default
error handlers.
"""
from functools import wraps
import http.client
import logging
import traceback
from typing import List

from flask import request

from zsl import inject
from zsl.db.model.app_model import AppModel
from zsl.errors import ErrorConfiguration, ErrorHandler, ErrorProcessor
from zsl.interface.task import ModelConversionError
from zsl.router.task import RoutingError
from zsl.task.job_context import JobContext, StatusCodeResponder, WebJobContext, add_responder
from zsl.task.task_decorator import ForbiddenError, json_output
from zsl.utils.documentation import documentation_link
from zsl.utils.http import get_http_status_code_value


class ErrorResponse(AppModel):
    def __init__(self, code, message):
        super().__init__({})
        self.code = code
        self.message = message


def register(e):
    # type: (ErrorHandler|ErrorProcessor)->None
    if isinstance(e, ErrorHandler):
        _error_handlers.append(e)
    if isinstance(e, ErrorProcessor):
        _error_processors.append(e)


class DefaultErrorHandler(ErrorHandler):
    ERROR_CODE = "UNKNOWN_ERROR"
    ERROR_MESSAGE = "An error occurred!"

    def can_handle(self, e):
        return True

    @json_output
    def handle(self, ex):
        logger = logging.getLogger(__name__)
        logger.error(str(ex) + "\n" + traceback.format_exc())
        logger.error("Request:\n{0}\n{1}\n".format(request.headers,
                                                   request.data))
        link = documentation_link('error_handling')
        logger.info("Provide your own error handler so that "
                    "a better error is produced, check {0}.".format(link))

        add_responder(StatusCodeResponder(http.client.INTERNAL_SERVER_ERROR))
        return ErrorResponse(self.ERROR_CODE, self.ERROR_MESSAGE)


class RoutingErrorHandler(ErrorHandler):
    ERROR_CODE = 'NOT_FOUND'

    def can_handle(self, e):
        return isinstance(e, RoutingError)

    @json_output
    def handle(self, ie):
        logging.error(str(ie) + "\n" + traceback.format_exc())
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.NOT_FOUND)))
        return ErrorResponse(self.ERROR_CODE, str(ie))


class ModelConversionErrorHandler(ErrorHandler):
    ERROR_CODE = 'INVALID_REQUEST'

    def can_handle(self, e):
        return isinstance(e, ModelConversionError)

    @json_output
    def handle(self, e):
        logging.error(str(e) + "\n" + traceback.format_exc())
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.UNPROCESSABLE_ENTITY)))
        return ErrorResponse(self.ERROR_CODE, str(e))


class ForbiddenErrorHandler(ErrorHandler):
    ERROR_CODE = 'FORBIDDEN'

    def can_handle(self, e):
        return isinstance(e, ForbiddenError)

    @json_output
    def handle(self, e):
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.FORBIDDEN)))
        return ErrorResponse(self.ERROR_CODE, str(e))


_DEFAULT_ERROR_HANDLER = DefaultErrorHandler()
_ROUTING_ERROR_HANDLER = RoutingErrorHandler()
_FORBIDDEN_ERROR_HANDLER = ForbiddenErrorHandler()
_MODEL_CONVERSION_ERROR_HANDLER = ModelConversionErrorHandler()
_error_handlers = [_MODEL_CONVERSION_ERROR_HANDLER, _FORBIDDEN_ERROR_HANDLER,
                   _ROUTING_ERROR_HANDLER]  # type: List[ErrorHandler]
_error_processors = []


def error_handler(f):
    """
    Default error handler.
     - On server side error shows a message
       'An error occurred!' and returns 500 status code.
     - Also serves well in the case when the resource/task/method
       is not found - returns 404 status code.
    """

    @wraps(f)
    def error_handling_function(*args, **kwargs):
        @inject(error_config=ErrorConfiguration)
        def get_error_configuration(error_config):
            # type:(ErrorConfiguration)->ErrorConfiguration
            return error_config

        def should_skip_handling():
            use_flask_handler = get_error_configuration().use_flask_handler
            is_web_request = isinstance(JobContext.get_current_context(), WebJobContext)
            return use_flask_handler and is_web_request

        try:
            return f(*args, **kwargs)
        except Exception as ex:
            if should_skip_handling():
                raise

            for ep in _error_processors:
                ep.handle(ex)

            for eh in _error_handlers:
                if eh.can_handle(ex):
                    return eh.handle(ex)

            return _DEFAULT_ERROR_HANDLER.handle(ex)

    return error_handling_function



from unittest import mock
from unittest.case import TestCase

import sqlalchemy
from sqlalchemy import inspect, text

from zsl import Injected, inject
from zsl.application.containers.container import IoCContainer
from zsl.application.modules.alchemy_module import AlchemyModule
from zsl.application.modules.context_module import DefaultContextModule
from zsl.application.modules.logger_module import LoggerModule
from zsl.application.modules.task_router import TaskRouterModule
from zsl.db.model.sql_alchemy import DeclarativeBase, metadata
from zsl.service import transactional
from zsl.service.service import TransactionalSupportMixin
from zsl.testing.db import IN_MEMORY_DB_SETTINGS, DatabaseSchemaInitializationException, DbTestCase, DbTestModule
from zsl.testing.db import TestSessionFactory as SessionFactoryForTesting
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class TestContainerNoTestSessionFactory(IoCContainer):
    logger = LoggerModule
    database = AlchemyModule
    context = DefaultContextModule
    task_router = TaskRouterModule


class TestContainerTestSessionFactory(TestContainerNoTestSessionFactory):
    session_factory = DbTestModule


class DbTestCase_SchemaInitializationAndQuerying_Test(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name="DbTestModuleTest",
        config_object=IN_MEMORY_DB_SETTINGS,
        container=TestContainerTestSessionFactory
    )

    # see `setUpClass` for definition of the Foo model
    _FOO_MODEL = None

    class DbTest1(DbTestCase, TransactionalSupportMixin, TestCase):

        @transactional
        def test_select_record_by_name_from_foo(self):
            return (
                self._orm.query(DbTestCase_SchemaInitializationAndQuerying_Test._FOO_MODEL)
                .filter(DbTestCase_SchemaInitializationAndQuerying_Test._FOO_MODEL.name == "test")
                .first()
            )

        @transactional
        def test_select_count_from_foo(self) -> int:
            return self._orm.query(DbTestCase_SchemaInitializationAndQuerying_Test._FOO_MODEL).count()

        @transactional
        def test_insert_into_foo_and_select_count(self):
            foo = DbTestCase_SchemaInitializationAndQuerying_Test._FOO_MODEL()
            foo.name = "test"
            self._orm.add(foo)
            self._orm.flush()
            count = self._orm.query(DbTestCase_SchemaInitializationAndQuerying_Test._FOO_MODEL).count()
            return count

    class DbTest2(DbTest1):
        pass

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # we are defining our Foo model in the setUpClass method in order to have control over Foo "registration" to
        # the global object `zsl.db.model.sql_alchemy.metadata`. We don't want the Foo model to be registered to the
        # global object outside the scope of this test class - it could cause problems in other tests.
        class Foo(DeclarativeBase):
            __tablename__ = 'foo'
            id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            name = sqlalchemy.Column(sqlalchemy.String, nullable=True)

        cls._FOO_MODEL = Foo

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        # removing our Foo model from the global object `zsl.db.model.sql_alchemy.metadata`
        metadata.remove(cls._FOO_MODEL.__table__)

        cls._FOO_MODEL = None

    @inject(engine=sqlalchemy.engine.Engine)
    def setUp(self, engine=Injected):
        super().setUp()
        # unfortunately we have to reset the db schema initialization to be able to test, if the schema is correctly
        # initialized. The reason is that the class `SessionFactoryForTesting` is also used in another tests and the
        # schema could already be initialized when running these tests.
        SessionFactoryForTesting.reset_db_schema_initialization(engine)

    @inject(engine=sqlalchemy.engine.Engine)
    def tearDown(self, engine=Injected):
        super().tearDown()
        SessionFactoryForTesting.reset_db_schema_initialization(engine)

    def test_db_schema_is_created_in_setupclass(self):
        test = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest1()
        with mock.patch('zsl.testing.db.metadata') as mock_metadata:
            test.setUpClass()

            mock_metadata.drop_all.assert_not_called()
            mock_metadata.create_all.assert_called_once_with(mock_metadata.bind)

    def test_db_schema_is_created_exactly_once_within_run_of_all_tests(self):
        test_1 = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest1()
        test_2 = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest2()
        with mock.patch('zsl.testing.db.metadata') as mock_metadata:
            test_1.setUpClass()
            test_1.setUp()
            test_1.tearDown()
            test_1.setUp()
            test_1.tearDown()
            test_1.tearDownClass()

            test_2.setUpClass()
            test_2.setUp()
            test_2.tearDown()
            test_2.tearDownClass()

            mock_metadata.drop_all.assert_not_called()
            mock_metadata.create_all.assert_called_once_with(mock_metadata.bind)

    @inject(engine=sqlalchemy.engine.Engine)
    def test__when_db_contains_table__then_exception_is_raised_within_setup_and_schema_is_not_created(
        self, engine
    ):
        # create a table in the database
        with engine.connect() as connection:
            connection.execute(text("CREATE TABLE bar (id INTEGER PRIMARY KEY, name TEXT)"))

        db_test = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest1()

        # database schema initialization should fail
        with self.assertRaises(DatabaseSchemaInitializationException):
            db_test.setUpClass()

        # check that no table was created or dropped
        inspector = inspect(engine)
        existing_tables = inspector.get_table_names()
        self.assertEqual(existing_tables, ["bar"], "No table should be dropped or created.")

        # clean up
        with engine.connect() as connection:
            connection.execute(text("DROP TABLE bar"))

    def test_tables_exists_within_tests(self):
        db_test_1 = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest1()
        db_test_2 = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest2()

        try:
            db_test_1.setUpClass()
            db_test_1.setUp()
            db_test_1.test_select_record_by_name_from_foo()
            db_test_1.tearDown()

            db_test_1.setUp()
            db_test_1.test_select_count_from_foo()
            db_test_1.tearDown()
            db_test_1.tearDownClass()

            db_test_2.setUpClass()
            db_test_2.setUp()
            db_test_2.test_select_count_from_foo()
            db_test_2.tearDown()
            db_test_2.tearDownClass()
        except Exception as e:
            self.fail("Exception raised: " + str(e))

    def test_data_created_within_test_are_not_available_in_another_test(self):
        db_test = DbTestCase_SchemaInitializationAndQuerying_Test.DbTest1()

        db_test.setUpClass()

        # inserting data into foo
        db_test.setUp()
        count_in_first_test = db_test.test_insert_into_foo_and_select_count()
        db_test.tearDown()

        db_test.setUp()
        # SELECT COUNT(*) FROM foo in another test
        count_in_second_test = db_test.test_select_count_from_foo()
        db_test.tearDown()

        db_test.tearDownClass()

        self.assertEqual(count_in_first_test, 1, "There should be one row in the table foo within first test")
        self.assertEqual(count_in_second_test, 0, "There should be no rows in the table foo within second test")
